import os

def linear_to_srgb(l):
	if l<0.0031308:
		return 12.92*l
	else:
		return 1.055*(l**(1/2.4))-0.055

def image_name(i):
	fp = i.filepath
	if fp:
		return os.path.split(fp)[1]
	else:
		return i.name


class ObjectExporter:
	def __init__(self):
		self.material_tex = False
		self.srgb_colors = True
		self.textures = "REF"
		self.separate_mesh = False
		self.separate_tech = False
		self.external_tech = True
		self.shared_tech = True

	def export(self, context, out_file, objs=None, progress=None):
		if objs is None:
			obj = context.active_object
		else:
			obj = objs[0]

		from .outfile import open_output
		out_file = open_output(out_file)

		from .export_mesh import MeshExporter
		mesh_export = MeshExporter()
		for k, v in self.__dict__.items():
			setattr(mesh_export, k, v)

		if self.separate_mesh:
			path, base = os.path.split(out_file.filename)
			base, ext = os.path.splitext(base)
			mesh_out = open_output(os.path.join(path, base+".mesh"))
			mesh = mesh_export.export(context, mesh_out, objs, progress)
			out_file.write("mesh", '"{}.mesh"'.format(base))
		else:
			out_file.begin("mesh")
			mesh = mesh_export.export(context, out_file, objs, progress)
			out_file.end()

		if self.external_tech and obj.technique:
			if obj.inherit_tech and mesh.materials[0].texture_slots:
				out_file.begin("technique")
				out_file.begin("inherit", '"{}"'.format(obj.technique))
				for slot in mesh.materials[0].texture_slots:
					if slot and slot.texture.type=="IMAGE":
						name = image_name(slot.texture.image)
						if slot.use_map_color_diffuse:
							out_file.write("texture", '"diffuse_map"', '"{}"'.format(name))
						elif slot.use_map_normal:
							out_file.write("texture", '"normal_map"', '"{}"'.format(name))
				out_file.end()
				out_file.end()
			else:
				out_file.write("technique", '"{}"'.format(obj.technique))
		elif self.separate_tech:
			path, base = os.path.split(out_file.filename)
			if self.shared_tech and mesh.materials:
				tech_name = mesh.materials[0].name+".tech"
			else:
				base, ext = os.path.splitext(base)
				tech_name = base+".tech"
			tech_out = open_output(os.path.join(path, tech_name))
			self.export_technique(mesh, tech_out)
			out_file.write("technique", '"{}"'.format(tech_name))
		else:
			out_file.begin("technique")
			self.export_technique(mesh, out_file)
			out_file.end()

	def export_technique(self, mesh, out_file):
		out_file.begin("pass", '""')
		if mesh.materials:
			if self.srgb_colors:
				cm = linear_to_srgb
			else:
				cm = lambda x: x

			if self.material_tex:
				out_file.begin("material")
				out_file.write("diffuse", 1.0, 1.0, 1.0, 1.0)
				out_file.end()
				index = 0
				for u in mesh.uv_layers:
					if u.name=="material_tex":
						index = u.unit
				out_file.begin("texunit", index)
				out_file.begin("texture2d")
				out_file.write("min_filter", "NEAREST")
				out_file.write("mag_filter", "NEAREST")
				out_file.write("storage", "RGB", len(mesh.materials), 1)
				texdata = '"'
				for m in mesh.materials:
					color = [int(cm(c)*255) for c in m.diffuse_color*mat.diffuse_intensity]
					texdata += "\\x%02X\\x%02X\\x%02X"%tuple(color)
				texdata += '"'
				out_file.write("raw_data", texdata)
				out_file.end()
				out_file.end()
			else:
				mat = mesh.materials[0]
				out_file.begin("material")
				if any((s and s.use_map_color_diffuse) for s in mat.texture_slots):
					out_file.write("diffuse", 1.0, 1.0, 1.0, 1.0)
					amb = cm(mat.ambient)
					out_file.write("ambient", amb, amb, amb, 1.0)
				else:
					diff = mat.diffuse_color*mat.diffuse_intensity
					out_file.write("diffuse", cm(diff.r), cm(diff.g), cm(diff.b), 1.0)
					amb = diff*mat.ambient
					out_file.write("ambient", cm(amb.r), cm(amb.g), cm(amb.b), 1.0)
				spec = mat.specular_color*mat.specular_intensity
				out_file.write("specular", spec.r, spec.g, spec.b, 1.0)
				out_file.write("shininess", mat.specular_hardness);
				out_file.end()

			if self.textures!="NONE":
				for slot in mesh.materials[0].texture_slots:
					if not slot:
						continue

					tex = slot.texture
					if tex.type!="IMAGE":
						continue

					if slot.uv_layer:
						for u in mesh.uv_layers:
							if u.name==slot.uv_layer:
								index = u.unit
					else:
						index = mesh.uv_layers[0].unit

					out_file.begin("texunit", index)
					if self.textures=="INLINE":
						out_file.begin("texture2d")
						out_file.write("min_filter", "LINEAR")
						out_file.write("storage", "RGBA", tex.image.size[0], tex.image.size[1])
						texdata = '"'
						for p in tex.image.pixels:
							texdata += "\\x%02X"%int(p*255)
						texdata += '"'
						out_file.write("raw_data", texdata)
						out_file.end()
					else:
						out_file.write("texture", '"%s"'%image_name(tex.image))
					out_file.end()

		out_file.end()
