#ifndef MSP_GL_FONT_H_
#define MSP_GL_FONT_H_

#include <map>
#include <string>
#include <msp/datafile/objectloader.h>
#include <msp/stringcodec/utf8.h>
#include "vertexarray.h"

namespace Msp {
namespace GL {

class PrimitiveBuilder;
class Texture2D;

/**
Stores a set of glyphs and creates strings out of them.
*/
class Font
{
public:
	class Loader: public DataFile::CollectionObjectLoader<Font>
	{
	public:
		Loader(Font &);
		Loader(Font &, Collection &);

	private:
		void init();
		void glyph(unsigned);
		void kerning(unsigned, unsigned, float);
		void texture();
		void texture_ref(const std::string &);
	};

	struct Glyph
	{
		class Loader: public Msp::DataFile::ObjectLoader<Glyph>
		{
		public:
			Loader(Glyph &);
		private:
			void texcoords(float, float, float, float);
		};

		unsigned code;
		float x1, y1;
		float x2, y2;
		float w, h;
		float off_x, off_y;
		float advance;
	};

private:
	typedef std::map<unsigned, Glyph> GlyphMap;
	typedef std::pair<unsigned, unsigned> KerningKey;
	typedef std::map<KerningKey, float> KerningMap;

	RefPtr<const Texture2D> texture;
	float native_size;
	float ascent;
	float descent;
	GlyphMap glyphs;
	KerningMap kerning;

public:
	Font();
	~Font();

	void set_texture(const Texture2D &);
	const Texture2D &get_texture() const;

	/** Adds a glyph to the font.  There must not be an existing glyph with the
	same code. */
	void add_glyph(const Glyph &);
	void set_kerning(unsigned, unsigned, float);

	/** Returns the size used to generate the font texture.  This serves as a
	hint for obtaining the best quality when rendering strings. */
	float get_native_size() const { return native_size; }

	float get_ascent() const { return ascent; }
	float get_descent() const { return descent; }

	/** Returns the width of a string, in multiples of the font size.  Scale the
	result according to the size used in rendering. */
	float get_string_width(const std::string &, StringCodec::Decoder &) const;

	template<class C>
	float get_string_width(const std::string &str) const
	{
		typename C::Decoder dec;
		return get_string_width(str, dec);
	}

	float get_string_width(const std::string &str) const
	{ return get_string_width<StringCodec::Utf8>(str); }

	/** Draws a string to the framebuffer with Immediate.  It is drawn with size
	1.0; set up matrices for the desired size before the call. */
	void draw_string(const std::string &, StringCodec::Decoder &, const Color & = Color()) const;

	template<class C>
	void draw_string(const std::string &str, const Color &color = Color()) const
	{
		typename C::Decoder dec;
		draw_string(str, dec, color);
	}

	void draw_string(const std::string &str, const Color &color = Color()) const
	{ draw_string<StringCodec::Utf8>(str, color); }

	/** Builds the primitives for a string.  Two-dimensional vertex and texture
	coordinates are generated.  Size 1.0 is used for building; set up the
	builder's matrix before the call.  The texture is not bound, to avoid
	unnecessary bindings when creating meshes. */
	void build_string(const std::string &, StringCodec::Decoder &, PrimitiveBuilder &) const;

	template<class C>
	void build_string(const std::string &str, PrimitiveBuilder &pb) const
	{
		typename C::Decoder dec;
		build_string(str, dec, pb);
	}

	void build_string(const std::string &str, PrimitiveBuilder &pb) const
	{ return build_string<StringCodec::Utf8>(str, pb); }

private:
	void create_glyph_vertices(const Glyph &, VertexBuilder &) const;
	float get_glyph_advance(unsigned, unsigned = 0) const;
};

} // namespace GL
} // namespace Msp

#endif
