#include <stdexcept>
#include "light.h"
#include "lighting.h"
#include "lightunit.h"
#include "matrix.h"
#include "misc.h"

using namespace std;

namespace Msp {
namespace GL {

Lighting::Lighting():
	ambient(0.2),
	sky_direction(0, 0, 1),
	horizon_angle(Geometry::Angle<float>::zero())
{ }

void Lighting::set_ambient(const Color &a)
{
	ambient = a;
}

void Lighting::set_sky_color(const Color &s)
{
	sky_color = s;
}

void Lighting::set_sky_direction(const Vector3 &d)
{
	sky_direction = d;
}

void Lighting::set_horizon_angle(const Geometry::Angle<float> &a)
{
	horizon_angle = a;
}

void Lighting::attach(unsigned i, const Light &l)
{
	if(i>=LightUnit::get_n_units())
		throw out_of_range("Lighting::attach");

	if(i>=lights.size())
		lights.resize(i+1);

	lights[i] = &l;
	if(current()==this)
		l.bind_to(i);
}

void Lighting::detach(unsigned i)
{
	if(i>=lights.size())
		return;

	lights[i] = 0;
	if(current()==this)
		Light::unbind_from(i);
}

void Lighting::update_shader_data(ProgramData &shdata, const Matrix &view_matrix) const
{
	shdata.uniform("ambient_color", ambient);
	shdata.uniform("sky_color", sky_color);
	shdata.uniform("eye_sky_dir", Vector3(view_matrix*Vector4(sky_direction, 0.0f)));
	shdata.uniform("horizon_limit", horizon_angle.radians());

	for(unsigned i=0; i<lights.size(); ++i)
		if(lights[i])
			lights[i]->update_shader_data(shdata, view_matrix, i);
}

void Lighting::bind() const
{
	if(!set_current(this))
		return;

	enable(GL_LIGHTING);
	glLightModelfv(GL_LIGHT_MODEL_AMBIENT, &ambient.r);
	for(unsigned i=0; i<lights.size(); ++i)
		if(lights[i])
			lights[i]->bind_to(i);
}

void Lighting::unbind()
{
	const Lighting *old = current();
	if(!set_current(0))
		return;

	for(unsigned i=0; i<old->lights.size(); ++i)
		if(old->lights[i])
			Light::unbind_from(i);

	disable(GL_LIGHTING);
}

} // namespace GL
} // namespace Msp
