#include <msp/datafile/collection.h>
#include <msp/strings/format.h>
#include "material.h"
#include "mesh.h"
#include "object.h"
#include "objectinstance.h"
#include "program.h"
#include "programdata.h"
#include "renderer.h"
#include "resourcemanager.h"
#include "technique.h"
#include "texturing.h"

using namespace std;

namespace Msp {
namespace GL {

Object::Object():
	meshes(1)
{ }

Object::Object(const Mesh *m, const Technique *t)
{
	set_mesh(m);
	set_technique(t);
}

// Avoid synthesizing ~RefPtr in files including object.h
Object::~Object()
{
	if(meshes[0])
		if(ResourceManager *rm = meshes[0]->get_manager())
			rm->unwatch_resource(*meshes[0], *this);
}

void Object::set_mesh(unsigned i, const Mesh *m)
{
	if(i>meshes.size())
		throw out_of_range("Object::set_mesh");

	if(i==meshes.size())
		meshes.push_back(m);
	else
	{
		if(i==0 && meshes[i])
			if(ResourceManager *rm = meshes[i]->get_manager())
				rm->unwatch_resource(*meshes[i], *this);
		meshes[i] = m;
	}
	meshes[i].keep();

	if(i==0 && m)
		if(ResourceManager *rm = m->get_manager())
			rm->watch_resource(*m, *this);

	update_bounding_sphere();
}

void Object::update_bounding_sphere()
{
	vector<Vector3> points;
	for(vector<RefPtr<const Mesh> >::const_iterator i=meshes.begin(); i!=meshes.end(); ++i)
	{
		const VertexArray &vertices = (*i)->get_vertices();

		int offset = vertices.get_format().offset(VERTEX3);
		bool three = true;
		if(offset<0)
		{
			offset = vertices.get_format().offset(VERTEX2);
			three = false;
			if(offset<0)
				continue;
		}

		unsigned n_vertices = vertices.size();
		points.reserve(points.size()+n_vertices);
		for(unsigned j=0; j<n_vertices; ++j)
		{
			const float *v = vertices[j];
			points.push_back(Vector3(v[offset], v[offset+1], (three ? v[offset+2] : 0.0f)));
		}
	}

	bounding_sphere = Geometry::BoundingSphere<float, 3>::from_point_cloud(points.begin(), points.end());
}

const Mesh *Object::get_mesh(unsigned i) const
{
	if(i>=meshes.size())
		return 0;

	return meshes[i].get();
}

void Object::set_technique(const Technique *t)
{
	technique = t;
	technique.keep();
}

void Object::render(const Tag &tag) const
{
	const RenderPass *pass = get_pass(tag);
	if(!pass)
		return;

	Bind bind_shader(pass->get_shader_program());
	if(pass->get_shader_data())
		pass->get_shader_data()->apply();
	Bind bind_material(pass->get_material());
	Bind bind_texturing(pass->get_texturing());

	meshes.front()->draw();
}

void Object::render(Renderer &renderer, const Tag &tag) const
{
	const RenderPass *pass = get_pass(tag);
	if(!pass)
		return;

	Renderer::Push push(renderer);
	pass->apply(renderer);

	setup_render(renderer, tag);
	meshes.front()->draw(renderer);
	finish_render(renderer, tag);
}

void Object::render(Renderer &renderer, const ObjectInstance &inst, const Tag &tag) const
{
	const RenderPass *pass = get_pass(tag);
	if(!pass)
		return;

	Renderer::Push push(renderer);
	pass->apply(renderer);

	setup_render(renderer, tag);
	inst.setup_render(renderer, tag);
	unsigned lod = min<unsigned>(inst.get_level_of_detail(renderer), meshes.size()-1);
	meshes[lod]->draw(renderer);
	inst.finish_render(renderer, tag);
	finish_render(renderer, tag);
}

const RenderPass *Object::get_pass(const Tag &tag) const
{
	if(!technique)
		throw logic_error("!technique");
	if(!technique->has_pass(tag))
		return 0;
	return &technique->get_pass(tag);
}

void Object::resource_loaded(Resource &res)
{
	if(!meshes.empty() && &res==meshes.front().get() && bounding_sphere.is_empty())
		update_bounding_sphere();
}


Object::Loader::Loader(Object &o):
	DataFile::CollectionObjectLoader<Object>(o, 0)
{
	init();
}

Object::Loader::Loader(Object &o, Collection &c):
	DataFile::CollectionObjectLoader<Object>(o, &c)
{
	init();
}

void Object::Loader::init()
{
	add("mesh",     &Loader::mesh_inline);
	add("mesh",     &Loader::mesh_inline_lod);
	add("mesh",     &Loader::mesh);
	add("mesh",     &Loader::mesh_lod);
	add("technique", &Loader::technique_inline);
	add("technique", &Loader::technique);
}

void Object::Loader::finish()
{
	obj.update_bounding_sphere();
}

void Object::Loader::mesh_inline()
{
	RefPtr<Mesh> msh = new Mesh;
	load_sub(*msh);
	obj.meshes.front() = msh;
}

void Object::Loader::mesh_inline_lod(unsigned l)
{
	if(l>obj.meshes.size())
		throw out_of_range("Object::Loader::mesh_inline_lod");

	RefPtr<Mesh> msh = new Mesh;
	load_sub(*msh);
	if(l==obj.meshes.size())
		obj.meshes.push_back(msh);
	else
		obj.meshes[l] = msh;
}

void Object::Loader::mesh(const std::string &n)
{
	obj.set_mesh(&get_collection().get<Mesh>(n));
}

void Object::Loader::mesh_lod(unsigned l, const string &n)
{
	obj.set_mesh(l, &get_collection().get<Mesh>(n));
}

void Object::Loader::technique_inline()
{
	RefPtr<Technique> tech = new Technique;
	if(coll)
		load_sub(*tech, get_collection());
	else
		load_sub(*tech);
	obj.technique = tech;
}

void Object::Loader::technique(const std::string &n)
{
	obj.set_technique(&get_collection().get<Technique>(n));
}

} // namespace GL
} // namespace Msp
