#ifndef MSP_GL_OBJECT_H_
#define MSP_GL_OBJECT_H_

#include <vector>
#include "bindable.h"
#include "renderable.h"
#include "renderpass.h"
#include "resourcewatcher.h"

namespace Msp {
namespace GL {

class Material;
class Mesh;
class ObjectInstance;
class Technique;
class Texture;

/**
Combines a Mesh with a Technique to give it an appearance.  The Technique will
define which render passes the Object supports.

In many cases, it's desirable to include multiple copies of an Object in a
Scene, with different model matrices.  ObjectInstances can be used to alter the
rendering of an object on a per-instance basis.
*/
class Object: public Renderable, private ResourceWatcher
{
public:
	class Loader: public DataFile::CollectionObjectLoader<Object>
	{
	public:
		Loader(Object &);
		Loader(Object &, Collection &);
	private:
		void init();
		virtual void finish();

		void mesh_inline();
		void mesh_inline_lod(unsigned);
		void mesh(const std::string &);
		void mesh_lod(unsigned, const std::string &);
		void technique_inline();
		void technique(const std::string &);
	};

private:
	std::vector<RefPtr<const Mesh> > meshes;
	RefPtr<const Technique> technique;
	Geometry::BoundingSphere<float, 3> bounding_sphere;

public:
	Object();
	Object(const Mesh *, const Technique *);
	~Object();

	void set_mesh(const Mesh *m) { set_mesh(0, m); }
	void set_mesh(unsigned, const Mesh *);
private:
	void update_bounding_sphere();
public:
	const Mesh *get_mesh(unsigned = 0) const;
	void set_technique(const Technique *);
	const Technique *get_technique() const { return technique.get(); }

	virtual const Geometry::BoundingSphere<float, 3> *get_bounding_sphere() const { return &bounding_sphere; }

	virtual void render(const Tag &tag = Tag()) const;

	virtual void render(Renderer &, const Tag & = Tag()) const;

	/** Renders an instance of the object.  The instance's hook functions are
	called before and after drawing the mesh. */
	virtual void render(Renderer &, const ObjectInstance &, const Tag & = Tag()) const;

protected:
	virtual void setup_render(Renderer &, const Tag &) const { }
	virtual void finish_render(Renderer &, const Tag &) const { }

private:
	const RenderPass *get_pass(const Tag &) const;

	virtual void resource_loaded(Resource &);
};

} // namespace GL
} // namespace Msp

#endif
