#include <algorithm>
#include <cstring>
#include <set>
#include <msp/core/hash.h>
#include <msp/core/maputils.h>
#include <msp/gl/extensions/arb_shader_objects.h>
#include <msp/gl/extensions/arb_uniform_buffer_object.h>
#include <msp/gl/extensions/arb_vertex_shader.h>
#include <msp/gl/extensions/ext_gpu_shader4.h>
#include <msp/strings/format.h>
#include "buffer.h"
#include "error.h"
#include "misc.h"
#include "program.h"
#include "shader.h"

using namespace std;

namespace Msp {
namespace GL {

Program::Program()
{
	init();
}

Program::Program(const ProgramBuilder::StandardFeatures &features)
{
	init();

	ProgramBuilder builder(features);
	builder.add_shaders(*this);
	link();
}

Program::Program(const string &vert, const string &frag)
{
	init();

	attach_shader_owned(new VertexShader(vert));
	attach_shader_owned(new FragmentShader(frag));
	link();
}

void Program::init()
{
	static Require _req(ARB_shader_objects);

	linked = false;
	id = glCreateProgram();
}

Program::~Program()
{
	for(ShaderList::iterator i=owned_data.begin(); i!=owned_data.end(); ++i)
		delete *i;
	glDeleteProgram(id);
}

void Program::attach_shader(Shader &shader)
{
	if(find(shaders.begin(), shaders.end(), &shader)==shaders.end())
	{
		glAttachShader(id, shader.get_id());
		shaders.push_back(&shader);
	}
}

void Program::attach_shader_owned(Shader *shader)
{
	attach_shader(*shader);
	if(find(owned_data.begin(), owned_data.end(), shader)==owned_data.end())
		owned_data.push_back(shader);
}

void Program::detach_shader(Shader &shader)
{
	ShaderList::iterator i = remove(shaders.begin(), shaders.end(), &shader);
	if(i!=shaders.end())
	{
		shaders.erase(i, shaders.end());
		glDetachShader(id, shader.get_id());
	}
}

void Program::bind_attribute(unsigned index, const string &name)
{
	static Require _req(ARB_vertex_shader);
	glBindAttribLocation(id, index, name.c_str());
}

void Program::bind_attribute(VertexComponent comp, const string &name)
{
	bind_attribute(get_component_type(comp), name);
}

void Program::bind_fragment_data(unsigned index, const string &name)
{
	static Require _req(EXT_gpu_shader4);
	glBindFragDataLocation(id, index, name.c_str());
}

void Program::link()
{
	for(ShaderList::iterator i=shaders.begin(); i!=shaders.end(); ++i)
		if(!(*i)->is_compiled())
			(*i)->compile();

	uniforms.clear();
	legacy_vars = false;

	glLinkProgram(id);
	linked = get_program_i(id, GL_LINK_STATUS);
	if(!linked)
		throw compile_error(get_info_log());

	unsigned count = get_program_i(id, GL_ACTIVE_UNIFORMS);
	vector<UniformInfo *> uniforms_by_index(count);
	for(unsigned i=0; i<count; ++i)
	{
		char name[128];
		int len = 0;
		int size;
		GLenum type;
		glGetActiveUniform(id, i, sizeof(name), &len, &size, &type, name);
		if(len && strncmp(name, "gl_", 3))
		{
			/* Some implementations report the first element of a uniform array,
			others report just the name of the array itself. */
			if(len>3 && !strcmp(name+len-3, "[0]"))
				name[len-3] = 0;

			UniformInfo &info = uniforms[name];
			info.block = 0;
			info.name = name;
			info.size = size;
			info.array_stride = 0;
			info.matrix_stride = 0;
			info.type = type;
			uniforms_by_index[i] = &info;
		}
		else
			legacy_vars = true;
	}

	count = get_program_i(id, GL_ACTIVE_ATTRIBUTES);
	for(unsigned i=0; i<count; ++i)
	{
		char name[128];
		int len = 0;
		int size;
		GLenum type;
		glGetActiveAttrib(id, i, sizeof(name), &len, &size, &type, name);
		if(len && !strncmp(name, "gl_", 3))
			legacy_vars = true;
	}

	if(ARB_uniform_buffer_object)
	{
		uniform_blocks.clear();

		std::set<unsigned> used_bind_points;
		count = get_program_i(id, GL_ACTIVE_UNIFORM_BLOCKS);
		for(unsigned i=0; i<count; ++i)
		{
			char name[128];
			int len;
			glGetActiveUniformBlockName(id, i, sizeof(name), &len, name);
			UniformBlockInfo &info = uniform_blocks[name];
			info.name = name;

			int value;
			glGetActiveUniformBlockiv(id, i, GL_UNIFORM_BLOCK_DATA_SIZE, &value);
			info.data_size = value;

			glGetActiveUniformBlockiv(id, i, GL_UNIFORM_BLOCK_ACTIVE_UNIFORMS, &value);
			vector<int> indices(value);
			glGetActiveUniformBlockiv(id, i, GL_UNIFORM_BLOCK_ACTIVE_UNIFORM_INDICES, &indices[0]);
			for(vector<int>::iterator j=indices.begin(); j!=indices.end(); ++j)
			{
				if(!uniforms_by_index[*j])
					throw logic_error("Program::link");
				info.uniforms.push_back(uniforms_by_index[*j]);
				uniforms_by_index[*j]->block = &info;
			}

			vector<unsigned> indices2(indices.begin(), indices.end());
			vector<int> values(indices.size());
			glGetActiveUniformsiv(id, indices.size(), &indices2[0], GL_UNIFORM_OFFSET, &values[0]);
			for(unsigned j=0; j<indices.size(); ++j)
				uniforms_by_index[indices[j]]->location = values[j];

			indices2.clear();
			for(vector<int>::iterator j=indices.begin(); j!=indices.end(); ++j)
				if(uniforms_by_index[*j]->size>1)
					indices2.push_back(*j);
			if(!indices2.empty())
			{
				glGetActiveUniformsiv(id, indices2.size(), &indices2[0], GL_UNIFORM_ARRAY_STRIDE, &values[0]);
				for(unsigned j=0; j<indices2.size(); ++j)
					uniforms_by_index[indices2[j]]->array_stride = values[j];
			}

			indices2.clear();
			for(vector<int>::iterator j=indices.begin(); j!=indices.end(); ++j)
			{
				GLenum t = uniforms_by_index[*j]->type;
				if(t==GL_FLOAT_MAT4 || t==GL_FLOAT_MAT3 || t==GL_FLOAT_MAT2 ||
					t==GL_FLOAT_MAT2x3 || t==GL_FLOAT_MAT2x4 || t==GL_FLOAT_MAT3x2 ||
					t==GL_FLOAT_MAT3x4 || t==GL_FLOAT_MAT4x2 || t==GL_FLOAT_MAT4x3)
					indices2.push_back(*j);
			}
			if(!indices2.empty())
			{
				glGetActiveUniformsiv(id, indices2.size(), &indices2[0], GL_UNIFORM_MATRIX_STRIDE, &values[0]);
				for(unsigned j=0; j<indices2.size(); ++j)
					uniforms_by_index[indices2[j]]->matrix_stride = values[j];
			}

			sort(info.uniforms.begin(), info.uniforms.end(), uniform_location_compare);
			info.layout_hash = compute_layout_hash(info.uniforms);
			unsigned n_bindings = BufferRange::get_n_uniform_buffer_bindings();
			info.bind_point = info.layout_hash%n_bindings;
			while(used_bind_points.count(info.bind_point))
				info.bind_point = (info.bind_point+1)%n_bindings;
			glUniformBlockBinding(id, i, info.bind_point);
			used_bind_points.insert(info.bind_point);
		}
	}

	UniformBlockInfo &default_block = uniform_blocks[string()];
	default_block.data_size = 0;
	default_block.bind_point = -1;

	for(UniformMap::iterator i=uniforms.begin(); i!=uniforms.end(); ++i)
		if(!i->second.block)
		{
			i->second.location = glGetUniformLocation(id, i->second.name.c_str());
			i->second.block = &default_block;
			default_block.uniforms.push_back(&i->second);
		}

	default_block.layout_hash = compute_layout_hash(default_block.uniforms);

	string layout_descriptor;
	for(UniformBlockMap::const_iterator i=uniform_blocks.begin(); i!=uniform_blocks.end(); ++i)
		layout_descriptor += format("%d:%x\n", i->second.bind_point, i->second.layout_hash);
	uniform_layout_hash = hash32(layout_descriptor);
}

Program::LayoutHash Program::compute_layout_hash(const vector<const UniformInfo *> &uniforms)
{
	string layout_descriptor;
	for(vector<const UniformInfo *>::const_iterator i = uniforms.begin(); i!=uniforms.end(); ++i)
		layout_descriptor += format("%d:%s:%x:%d\n", (*i)->location, (*i)->name, (*i)->type, (*i)->size);
	return hash32(layout_descriptor);
}

bool Program::uniform_location_compare(const UniformInfo *uni1, const UniformInfo *uni2)
{
	return uni1->location<uni2->location;
}

string Program::get_info_log() const
{
	GLsizei len = get_program_i(id, GL_INFO_LOG_LENGTH);
	char *buf = new char[len+1];
	glGetProgramInfoLog(id, len+1, &len, buf);
	string log(buf, len);
	delete[] buf;
	return log;
}

const Program::UniformBlockInfo &Program::get_uniform_block_info(const string &name) const
{
	return get_item(uniform_blocks, name);
}

const Program::UniformInfo &Program::get_uniform_info(const string &name) const
{
	return get_item(uniforms, name);
}

int Program::get_uniform_location(const string &n) const
{
	UniformMap::const_iterator i = uniforms.find(n);
	if(i==uniforms.end())
	{
		if(n[n.size()-1]==']')
		{
			string::size_type open_bracket = n.rfind('[');
			if(open_bracket!=string::npos)
			{
				/* The requested name looks like an array.  glGetActiveUniform only
				gives us the first element of the array, so try to look that up and
				add an offset. */
				unsigned offset = lexical_cast<unsigned>(n.substr(open_bracket+1, n.size()-2-open_bracket));
				i = uniforms.find(n.substr(0, open_bracket));
				if(i!=uniforms.end() && i->second.block->bind_point<0 && offset<i->second.size)
					return i->second.location+offset;
			}
		}
		return -1;
	}

	return i->second.block->bind_point<0 ? i->second.location : -1;
}

void Program::bind() const
{
	if(!linked)
		throw invalid_operation("Program::bind");

	if(!set_current(this))
		return;

	glUseProgram(id);
}

void Program::unbind()
{
	if(!set_current(0))
		return;

	glUseProgram(0);
}


Program::Loader::Loader(Program &p):
	DataFile::ObjectLoader<Program>(p)
{
	add("attribute",       &Loader::attribute);
	add("fragment_shader", &Loader::fragment_shader);
	add("standard",        &Loader::standard);
	add("vertex_shader",   &Loader::vertex_shader);
}

void Program::Loader::finish()
{
	obj.link();
}

void Program::Loader::attribute(unsigned i, const string &n)
{
	obj.bind_attribute(i, n);
}

void Program::Loader::fragment_shader(const string &src)
{
	obj.attach_shader_owned(new FragmentShader(src));
}

void Program::Loader::standard()
{
	ProgramBuilder::StandardFeatures feat;
	load_sub(feat);
	ProgramBuilder builder(feat);
	builder.add_shaders(obj);
}

void Program::Loader::vertex_shader(const string &src)
{
	obj.attach_shader_owned(new VertexShader(src));
}

} // namespace GL
} // namespace Msp
