#ifndef MSP_GL_PROGRAMDATA_H_
#define MSP_GL_PROGRAMDATA_H_

#include <map>
#include <msp/datafile/objectloader.h>
#include "program.h"
#include "vector.h"

namespace Msp {
namespace GL {

class Buffer;
class Matrix;
class Uniform;
class UniformBlock;
struct Color;

/**
Stores uniform variables for shader programs.  The uniforms are stored in a
program-independent way, and UniformBlocks are created to match the uniform
layouts of different programs.  If multiple programs have the same layout, the
same block is used for them.

The class is optimized for an access pattern where the set of uniforms and
programs stays constants, with only the values changing.
*/
class ProgramData
{
public:
	class Loader: public DataFile::ObjectLoader<ProgramData>
	{
	public:
		Loader(ProgramData &);
	private:
		void uniform1i(const std::string &, int);
		void uniform1f(const std::string &, float);
		void uniform2f(const std::string &, float, float);
		void uniform3f(const std::string &, float, float, float);
		void uniform4f(const std::string &, float, float, float, float);
	};

private:
	typedef unsigned Mask;

	enum
	{
		MASK_BITS = sizeof(Mask)*8,
		ALL_ONES = static_cast<Mask>(-1)
	};

	struct SharedBlock
	{
		Mask used;
		Mask dirty;
		UniformBlock *block;

		SharedBlock();
		SharedBlock(unsigned, UniformBlock *);
	};

	struct ProgramBlock
	{
		int bind_point;
		UniformBlock *block;
		SharedBlock *shared;

		ProgramBlock();
		ProgramBlock(int, SharedBlock *);
	};

	struct ProgramUniforms
	{
		std::vector<ProgramBlock> blocks;
		Mask used;
		Mask dirty;

		ProgramUniforms();
	};

	typedef std::map<std::string, unsigned> SlotMap;
	typedef std::map<Program::LayoutHash, SharedBlock> BlockMap;
	typedef std::map<Program::LayoutHash, ProgramUniforms> ProgramMap;

	// XXX All these mutables are a bit silly, but I'm out of better ideas
	SlotMap uniform_slots;
	std::vector<Uniform *> uniforms;
	mutable BlockMap blocks;
	mutable ProgramMap programs;
	mutable UniformBlock *last_block;
	mutable Buffer *buffer;
	mutable unsigned dirty;

public:
	ProgramData();
	ProgramData(const ProgramData &);
	ProgramData &operator=(const ProgramData &);
	~ProgramData();

private:
	void uniform(const std::string &, Uniform *);
public:
	void uniform(const std::string &, int);
	void uniform(const std::string &, float);
	void uniform(const std::string &, float, float);
	void uniform2(const std::string &, const float *);
	void uniform(const std::string &, float, float, float);
	void uniform(const std::string &, const Vector3 &);
	void uniform3(const std::string &, const float *);
	void uniform(const std::string &, float, float, float, float);
	void uniform(const std::string &, const Vector4 &);
	void uniform(const std::string &, const Color &);
	void uniform4(const std::string &, const float *);
	void uniform_matrix2(const std::string &, const float *);
	void uniform_matrix3(const std::string &, const float *);
	void uniform(const std::string &, const Matrix &);
	void uniform_matrix4(const std::string &, const float *);
	void uniform1_array(const std::string &, unsigned, const float *);
	void uniform2_array(const std::string &, unsigned, const float *);
	void uniform3_array(const std::string &, unsigned, const float *);
	void uniform4_array(const std::string &, unsigned, const float *);
	void uniform_matrix4_array(const std::string &, unsigned, const float *);

private:
	unsigned compute_slot_mask(const Program::UniformBlockInfo &) const;
	void update_block(UniformBlock &, const Program::UniformBlockInfo &) const;
	SharedBlock *get_shared_block(const Program::UniformBlockInfo &) const;

public:
	/** Applies uniform blocks for the currently bound program, creating them
	if needed. */
	void apply() const;
};

} // namespace GL
} // namespace Msp

#endif
