#include "batch.h"
#include "buffer.h"
#include "camera.h"
#include "error.h"
#include "lighting.h"
#include "material.h"
#include "mesh.h"
#include "program.h"
#include "programdata.h"
#include "renderable.h"
#include "renderer.h"
#include "texture.h"
#include "texturing.h"
#include "texunit.h"
#include "vertexarray.h"
#include "windingtest.h"

using namespace std;

namespace Msp {
namespace GL {

Renderer::Renderer(const Camera *c):
	mtx_stack(*this),
	mtx_changed(true),
	matrices_loaded(false),
	camera(c),
	state_stack(1),
	lighting_changed(false),
	element_buffer(0)
{
	state_stack.reserve(16);
	shdata_stack.reserve(32);
	state = &state_stack.back();

	if(camera)
	{
		mtx_stack.load(camera->get_matrix());
		standard_shdata.uniform("projection_matrix", camera->get_projection_matrix());
	}
	else
	{
		standard_shdata.uniform("projection_matrix", MatrixStack::projection().top());
		mtx_stack.load(MatrixStack::modelview().top());
	}
}

Renderer::~Renderer()
{
	if(matrices_loaded)
	{
		if(camera)
			MatrixStack::projection().pop();
		MatrixStack::modelview().pop();
	}

	Mesh::unbind();
	Texturing::unbind();
	Texture::unbind_from(0);
	Material::unbind();
	Lighting::unbind();
	Program::unbind();
	Buffer::unbind_from(ELEMENT_ARRAY_BUFFER);
	WindingTest::unbind();
}

void Renderer::set_texture(const Texture *t)
{
	state->texture = t;
	state->texturing = 0;
}

void Renderer::set_texturing(const Texturing *t)
{
	state->texturing = t;
	state->texture = 0;
}

unsigned Renderer::allocate_effect_texunit()
{
	return --state->lowest_effect_texunit;
}

void Renderer::set_material(const Material *m)
{
	state->material = m;
}

void Renderer::set_lighting(const Lighting *l)
{
	state->lighting = l;
	state->lighting_matrix = mtx_stack.top();
	if(l)
		l->update_shader_data(standard_shdata, mtx_stack.top());
	lighting_changed = true;
}

void Renderer::set_shader_program(const Program *p, const ProgramData *d)
{
	state->shprog = p;
	if(p && d)
		add_shader_data(*d);

	/* Even if we have no new shdata, the existing ones need to be re-applied
	to the new program */
	shdata_changed = true;
}

void Renderer::add_shader_data(const ProgramData &d)
{
	shdata_stack.push_back(&d);
	state->shdata_count = shdata_stack.size();
	shdata_changed = true;
}

void Renderer::set_mesh(const Mesh *m)
{
	state->mesh = m;
}

void Renderer::set_element_buffer(const Buffer *b)
{
	element_buffer = b;
}

void Renderer::set_winding_test(const WindingTest *w)
{
	state->winding_test = w;
}

void Renderer::set_reverse_winding(bool r)
{
	state->reverse_winding = r;
}

void Renderer::push_state()
{
	state_stack.push_back(state_stack.back());
	state = &state_stack.back();
	mtx_stack.push();
}

void Renderer::pop_state()
{
	if(state_stack.size()==1)
		throw stack_underflow("Renderer::pop_state");

	state_stack.pop_back();
	state = &state_stack.back();
	if(shdata_stack.size()>state->shdata_count)
		shdata_stack.erase(shdata_stack.begin()+state->shdata_count, shdata_stack.end());
	mtx_stack.pop();
	mtx_changed = true;
	shdata_changed = true;
}

void Renderer::escape()
{
	apply_state();
	Buffer::unbind_from(ELEMENT_ARRAY_BUFFER);
	matrices_loaded = false;
}

void Renderer::exclude(const Renderable &renderable)
{
	excluded.insert(&renderable);
}

void Renderer::include(const Renderable &renderable)
{
	excluded.erase(&renderable);
}

void Renderer::render(const Renderable &renderable, const Tag &tag)
{
	if(!excluded.count(&renderable))
		renderable.render(*this, tag);
}

void Renderer::draw(const Batch &batch)
{
	apply_state();

	bool legacy_bindings = (!state->shprog || state->shprog->uses_legacy_variables());
	if(legacy_bindings)
	{
		if(element_buffer)
			element_buffer->bind_to(ELEMENT_ARRAY_BUFFER);
		else
			Buffer::unbind_from(ELEMENT_ARRAY_BUFFER);
	}

	batch.draw();
}

void Renderer::apply_state()
{
	/* We (mostly) let the objects themselves figure out if the binding has
	changed */

	bool legacy_bindings = (!state->shprog || state->shprog->uses_legacy_variables());

	if(state->texturing)
		state->texturing->bind();
	else
	{
		Texturing::unbind();
		if(state->texture)
			state->texture->bind_to(0);
		else
			Texture::unbind_from(0);
	}

	if(legacy_bindings)
	{
		if(state->material)
			state->material->bind();
		else
			Material::unbind();

		if(lighting_changed)
		{
			if(state->lighting)
			{
				MatrixStack::modelview() = state->lighting_matrix;
				state->lighting->bind();
				mtx_changed = true;
				lighting_changed = false;
			}
			else
				Lighting::unbind();
		}
	}

	if(state->shprog)
	{
		state->shprog->bind();

		if(!legacy_bindings)
		{
			const Matrix &m = mtx_stack.top();
			standard_shdata.uniform("eye_obj_matrix", mtx_stack.top());
			LinAl::SquareMatrix<float, 3> nm;
			for(unsigned i=0; i<3; ++i)
				for(unsigned j=0; j<3; ++j)
					nm(i, j) = m(i, j);
			nm = transpose(invert(nm));
			standard_shdata.uniform_matrix3("eye_obj_normal_matrix", &nm(0, 0));
			if(state->material)
				state->material->get_shader_data().apply();
			standard_shdata.apply();
		}

		if(shdata_changed)
		{
			for(vector<const ProgramData *>::const_iterator i=shdata_stack.begin(); i!=shdata_stack.end(); ++i)
				(*i)->apply();
			shdata_changed = false;
		}
	}
	else
		Program::unbind();

	if(state->mesh)
	{
		if(legacy_bindings)
		{
			Mesh::unbind();
			state->mesh->get_vertices().apply();
		}
		else
			state->mesh->bind();
	}
	else
		Mesh::unbind();

	if(state->winding_test)
	{
		if(state->reverse_winding)
			state->winding_test->get_reverse().bind();
		else
			state->winding_test->bind();
	}
	else
		WindingTest::unbind();

	if(legacy_bindings)
	{
		if(!matrices_loaded)
		{
			MatrixStack::modelview().push();
			if(camera)
			{
				MatrixStack::projection().push();
				camera->apply();
			}
			matrices_loaded = true;
		}

		if(mtx_changed)
		{
			MatrixStack::modelview() = mtx_stack.top();
			mtx_changed = false;
		}
	}
}


Renderer::State::State():
	texture(0),
	texturing(0),
	lowest_effect_texunit(TexUnit::get_n_units()),
	material(0),
	lighting(0),
	shprog(0),
	shdata_count(0),
	mesh(0),
	winding_test(0),
	reverse_winding(false)
{ }


Renderer::MtxStack::MtxStack(Renderer &r):
	renderer(r)
{ }

void Renderer::MtxStack::update()
{
	renderer.mtx_changed = true;
}

} // namespace GL
} // namespace Msp
