#include <cmath>
#include <cstdlib>
#include "camera.h"
#include "light.h"
#include "matrix.h"
#include "misc.h"
#include "renderer.h"
#include "scene.h"
#include "shadowmap.h"
#include "tests.h"
#include "texunit.h"

using namespace std;

namespace Msp {
namespace GL {

ShadowMap::ShadowMap(unsigned s, const Renderable &r, const Light &l):
	Effect(r),
	size(s),
	light(l),
	radius(1),
	depth_bias(4),
	rendered(false)
{
	depth_buf.set_min_filter(LINEAR);
	depth_buf.set_compare_enabled(true);
	depth_buf.set_compare_func(LEQUAL);
	depth_buf.set_wrap(CLAMP_TO_EDGE);
	depth_buf.storage(DEPTH_COMPONENT, size, size);
	fbo.attach(DEPTH_ATTACHMENT, depth_buf, 0);
	fbo.require_complete();

	set_darkness(0.7);
}

void ShadowMap::set_target(const Vector3 &t, float r)
{
	target = t;
	radius = r;
}

void ShadowMap::set_darkness(float d)
{
	if(d<0.0f || d>1.0f)
		throw invalid_argument("ShadowMap::set_darkness");

	shdata.uniform("shadow_darkness", d);
}

void ShadowMap::set_depth_bias(float b)
{
	if(b<0.0f)
		throw invalid_argument("ShadowMap::set_depth_bias");

	depth_bias = b;
}

void ShadowMap::setup_frame() const
{
	if(rendered)
		return;

	renderable.setup_frame();
	rendered = true;

	const Vector4 &lpos = light.get_position();
	Vector3 back;
	if(lpos.w)
	{
		/* XXX Not really proper way to support positional lights, but good
		enough when the light source is far away */
		back = Vector3(lpos)-target;
	}
	else
		back = Vector3(lpos);
	back.normalize();

	Vector3 up;
	if(abs(back.z)<0.99)
		up = Vector3(0, 0, 1);
	else
		up = Vector3(0, 1, 0);

	Vector3 right = normalize(cross(up, back));

	Vector4 columns[4];
	columns[0] = Vector4(right, 0.0f);
	columns[1] = Vector4(normalize(cross(back, right)), 0.0f);
	columns[2] = Vector4(back, 0.0f);
	columns[3] = Vector4(target, 1.0f);
	light_matrix = Matrix::from_columns(columns);
	view_matrix = invert(light_matrix);

	MatrixStack::Push push_mv(MatrixStack::modelview());
	MatrixStack::Push push_proj(MatrixStack::projection());

	MatrixStack::projection() = Matrix::ortho(-radius, radius, -radius, radius, -radius, radius);
	MatrixStack::modelview() = view_matrix;

	shadow_matrix = light_matrix;
	shadow_matrix.scale(radius*2, radius*2, -radius*2);
	shadow_matrix.translate(-0.5, -0.5, depth_bias/size-0.5);
	shadow_matrix.invert();

	BindRestore bind_fbo(fbo);
	Bind bind_depth(DepthTest::lequal());
	fbo.clear(DEPTH_BUFFER_BIT);
	renderable.render("shadow");
}

void ShadowMap::finish_frame() const
{
	renderable.finish_frame();
	rendered = false;
}

void ShadowMap::render(Renderer &renderer, const Tag &tag) const
{
	if(!enabled_passes.count(tag))
		return renderer.render(renderable, tag);

	Renderer::Push _push_rend(renderer);

	unsigned unit = renderer.allocate_effect_texunit();
	int iunit = unit;
	shdata.uniform("shadow", iunit);

	Bind _bind_depth(depth_buf, unit);

	if(const Camera *camera = renderer.get_camera())
		/* Multiply by camera's object matrix to form a matrix that transforms
		from eye space to shadow space. */
		shdata.uniform("shd_eye_matrix", shadow_matrix*camera->get_object_matrix());
	else
		shdata.uniform("shd_eye_matrix", shadow_matrix);

	renderer.add_shader_data(shdata);
	renderer.render(renderable, tag);
}

} // namespace GL
} // namespace Msp
