#include <msp/io/memory.h>
#include "bindable.h"
#include "buffer.h"
#include "error.h"
#include "pixelstore.h"
#include "resources.h"
#include "texture2d.h"

using namespace std;

namespace Msp {
namespace GL {

class Texture2D::AsyncLoader: public Resource::AsyncLoader
{
private:
	Texture2D &texture;
	IO::Seekable &io;
	Buffer pixel_buffer;
	char *mapped_address;
	Graphics::Image image;
	unsigned n_bytes;
	int phase;

public:
	AsyncLoader(Texture2D &, IO::Seekable &);

	virtual bool needs_sync() const;
	virtual bool process();
};


Texture2D::Texture2D(ResourceManager *m):
	Texture(GL_TEXTURE_2D, m),
	ifmt(RGB),
	width(0),
	height(0),
	allocated(0)
{ }

void Texture2D::storage(PixelFormat fmt, unsigned wd, unsigned ht)
{
	if(width>0)
		throw invalid_operation("Texture2D::storage");
	if(wd==0 || ht==0)
		throw invalid_argument("Texture2D::storage");
	require_pixelformat(fmt);

	ifmt = fmt;
	width = wd;
	height = ht;
}

void Texture2D::allocate(unsigned level)
{
	if(allocated&(1<<level))
		return;

	image(level, get_base_pixelformat(ifmt), UNSIGNED_BYTE, 0);
}

void Texture2D::image(unsigned level, PixelFormat fmt, DataType type, const void *data)
{
	if(width==0 || height==0)
		throw invalid_operation("Texture2D::image");

	unsigned w = width;
	unsigned h = height;
	get_level_size(level, w, h);

	BindRestore _bind(this);
	glTexImage2D(target, level, ifmt, w, h, 0, fmt, type, data);

	allocated |= 1<<level;
	if(gen_mipmap && level==0)
	{
		for(; (w || h); w>>=1, h>>=1, ++level) ;
		allocated |= (1<<level)-1;
	}
}

void Texture2D::sub_image(unsigned level, int x, int y, unsigned wd, unsigned ht, PixelFormat fmt, DataType type, const void *data)
{
	if(width==0 || height==0)
		throw invalid_operation("Texture2D::sub_image");

	allocate(level);

	BindRestore _bind(this);
	glTexSubImage2D(target, level, x, y, wd, ht, fmt, type, data);
}

void Texture2D::load_image(const string &fn, bool srgb)
{
	Graphics::Image img;
	img.load_file(fn);

	image(img, srgb);
}

void Texture2D::image(const Graphics::Image &img, bool srgb)
{
	image(img, srgb, false);
}

void Texture2D::image(const Graphics::Image &img, bool srgb, bool from_buffer)
{
	unsigned w = img.get_width();
	unsigned h = img.get_height();
	PixelFormat fmt = pixelformat_from_graphics(img.get_format());
	if(width==0)
	{
		PixelFormat f = storage_pixelformat_from_graphics(img.get_format());
		if(srgb)
			f = get_srgb_pixelformat(f);
		storage(f, w, h);
	}
	else if(w!=width || h!=height)
		throw incompatible_data("Texture2D::image");

	PixelStore pstore = PixelStore::from_image(img);
	BindRestore _bind_ps(pstore);

	image(0, fmt, UNSIGNED_BYTE, from_buffer ? 0 : img.get_data());
}

void Texture2D::get_level_size(unsigned level, unsigned &w, unsigned &h)
{
	w >>= level;
	h >>= level;

	if(!w && h)
		w = 1;
	else if(!h && w)
		h = 1;
}

Resource::AsyncLoader *Texture2D::load(IO::Seekable &io)
{
	return new AsyncLoader(*this, io);
}

UInt64 Texture2D::get_data_size() const
{
	return id ? width*height*get_component_count(ifmt) : 0;
}

void Texture2D::unload()
{
	glDeleteTextures(1, &id);
	id = 0;
	// TODO check which params actually need refreshing
	dirty_params = -1;
}


Texture2D::Loader::Loader(Texture2D &t):
	DataFile::DerivedObjectLoader<Texture2D, Texture::Loader>(t)
{
	init();
}

Texture2D::Loader::Loader(Texture2D &t, Collection &c):
	DataFile::DerivedObjectLoader<Texture2D, Texture::Loader>(t, c)
{
	init();
}

void Texture2D::Loader::init()
{
	add("image_data", &Loader::image_data);
	add("raw_data", &Loader::raw_data);
	add("storage", &Loader::storage);
	add("storage", &Loader::storage_b);
}

void Texture2D::Loader::image_data(const string &data)
{
	Graphics::Image img;
	IO::Memory mem(data.data(), data.size());
	img.load_io(mem);

	obj.image(img, srgb);
}

void Texture2D::Loader::raw_data(const string &data)
{
	obj.image(0, get_base_pixelformat(obj.ifmt), UNSIGNED_BYTE, data.data());
}

void Texture2D::Loader::storage(PixelFormat fmt, unsigned w, unsigned h)
{
	if(srgb)
		fmt = get_srgb_pixelformat(fmt);
	obj.storage(fmt, w, h);
}

void Texture2D::Loader::storage_b(PixelFormat fmt, unsigned w, unsigned h, unsigned)
{
	storage(fmt, w, h);
}


Texture2D::AsyncLoader::AsyncLoader(Texture2D &t, IO::Seekable &i):
	texture(t),
	io(i),
	pixel_buffer(PIXEL_UNPACK_BUFFER),
	mapped_address(0),
	phase(0)
{ }

bool Texture2D::AsyncLoader::needs_sync() const
{
	return phase%2;
}

bool Texture2D::AsyncLoader::process()
{
	if(phase==0)
	{
		/* TODO Enhance the ImageLoader system so that the image can be loaded
		directly to the buffer */
		image.load_io(io);
		n_bytes = image.get_stride()*image.get_height();
	}
	else if(phase==1)
	{
		pixel_buffer.data(n_bytes, 0);
		mapped_address = reinterpret_cast<char *>(pixel_buffer.map(WRITE_ONLY));
	}
	else if(phase==2)
	{
		const char *data = reinterpret_cast<const char *>(image.get_data());
		copy(data, data+n_bytes, mapped_address);
	}
	else if(phase==3)
	{
		Bind _bind_buf(pixel_buffer, PIXEL_UNPACK_BUFFER);
		if(!pixel_buffer.unmap())
		{
			phase = 1;
			return false;
		}

		if(!texture.id)
			glGenTextures(1, &texture.id);
		texture.image(image, false, true);
	}

	++phase;
	return phase>3;
}

} // namespace GL
} // namespace Msp
