#include <cmath>
#include <msp/gl/extensions/ext_texture3d.h>
#include <msp/graphics/image.h>
#include "bindable.h"
#include "error.h"
#include "pixelstore.h"
#include "texture3d.h"

using namespace std;

namespace Msp {
namespace GL {

Texture3D::Texture3D():
	Texture(GL_TEXTURE_3D),
	ifmt(RGB),
	width(0),
	height(0),
	depth(0),
	allocated(0)
{
	static Require _req(EXT_texture3D);
}

void Texture3D::storage(PixelFormat fmt, unsigned wd, unsigned ht, unsigned dp)
{
	if(width>0)
		throw invalid_operation("Texture3D::storage");
	if(wd==0 || ht==0 || dp==0)
		throw invalid_argument("Texture3D::storage");
	require_pixelformat(fmt);

	width = wd;
	height = ht;
	depth = dp;
	ifmt = fmt;
}

void Texture3D::allocate(unsigned level)
{
	if(allocated&(1<<level))
		return;

	image(level, get_base_pixelformat(ifmt), UNSIGNED_BYTE, 0);
}

void Texture3D::image(unsigned level, PixelFormat fmt, DataType type, const void *data)
{
	if(width==0 || height==0 || depth==0)
		throw invalid_operation("Texture3D::image");

	unsigned w = width;
	unsigned h = height;
	unsigned d = depth;
	get_level_size(level, w, h, d);

	BindRestore _bind(this);
	glTexImage3D(target, level, ifmt, width, height, depth, 0, fmt, type, data);

	allocated |= 1<<level;
	if(gen_mipmap && level==0)
	{
		for(; (w || h || d); w>>=1, h>>=1, d>>=1, ++level) ;
		allocated |= (1<<level)-1;
	}
}

void Texture3D::sub_image(unsigned level, int x, int y, int z, unsigned wd, unsigned ht, unsigned dp, PixelFormat fmt, DataType type, const void *data)
{
	if(width==0 || height==0 || depth==0)
		throw invalid_operation("Texture3D::image");

	allocate(level);

	BindRestore _bind(this);
	glTexSubImage3D(target, level, x, y, z, wd, ht, dp, fmt, type, data);
}

void Texture3D::load_image(const string &fn, int dp)
{
	Graphics::Image img;
	img.load_file(fn);

	unsigned w = img.get_width();
	unsigned h = img.get_height();
	unsigned d = 1;

	if(dp==-1)
	{
		if(h%w)
			throw incompatible_data("Texture3D::load_image");
		d = h/w;
		h = w;
	}
	else if(dp==-2)
	{
		for(d=h; d*d>h; d>>=2) ;
		for(; d*d<h; ++d) ;
		if(d*d!=h)
			throw incompatible_data("Texture3D::load_image");
		h = d;
	}
	else if(dp>0)
	{
		d = dp;
		if(h%d)
			throw incompatible_data("Texture3D::load_image");
		h /= d;
	}
	else
		throw invalid_argument("Texture3D::load_image");

	PixelFormat fmt = pixelformat_from_graphics(img.get_format());
	if(width==0)
		storage(storage_pixelformat_from_graphics(img.get_format()), w, h, d);
	else if(w!=width || h!=height || d!=depth)
		throw incompatible_data("Texture3D::load_image");

	PixelStore pstore = PixelStore::from_image(img);
	BindRestore _bind_ps(pstore);

	image(0, fmt, UNSIGNED_BYTE, img.get_data());
}

void Texture3D::get_level_size(unsigned level, unsigned &w, unsigned &h, unsigned &d)
{
	w >>= level;
	h >>= level;
	d >>= level;

	if(!w && (h || d))
		w = 1;
	if(!h && (w || d))
		h = 1;
	if(!d && (w || h))
		d = 1;
}

UInt64 Texture3D::get_data_size() const
{
	return id ? width*height*depth*get_pixel_size(ifmt) : 0;
}

} // namespace GL
} // namespace Msp
