#ifndef MSP_GL_TEXTURE3D_H_
#define MSP_GL_TEXTURE3D_H_

#include <string>
#include "datatype.h"
#include "pixelformat.h"
#include "texture.h"

namespace Msp {
namespace GL {

/**
Three-dimensional texture.  Consists of an array of texels in the shape of a
right cuboid.  Texture coordinates have a principal range of [0, 1].
*/
class Texture3D: public Texture
{
private:
	PixelFormat ifmt;
	unsigned width;
	unsigned height;
	unsigned depth;
	unsigned allocated;

public:
	Texture3D();

	/** Defines storage structure for the texture.  Must be called before an
	image can be uploaded.  Once storage is defined, it can't be changed. */
	void storage(PixelFormat fmt, unsigned wd, unsigned ht, unsigned dp);

	/** Allocates storage for the texture.  The contents are initially
	undefined.  If storage has already been allocated, does nothing. */
	void allocate(unsigned level);

	/** Uploads an image to the texture.  Storage must be defined beforehand.
	The image data must have dimensions and format compatible with the defined
	storage. */
	void image(unsigned level, PixelFormat fmt, DataType type, const void *data);

	/** Updates a cuboid-shaped region of the texture.  Storage must be defined
	and allocated beforehand.  The update region must be fully inside the
	texture. */
	void sub_image(unsigned level,
		int x, int y, int z, unsigned wd, unsigned ht, unsigned dp,
		PixelFormat fmt, DataType type, const void *data);

	/** Loads an image from a file and uploads it to the texture.  If storage
	has not been defined, it will be set to match the loaded image.  To
	construct a three-dimensional texture from a two-dimensional image, the
	image is interpreted as an array of consecutive images.  If dp is -1, the
	texture's width and height are equal.  If dp is -2, the texture's height and
	depth are equal.  Otherwise, dp must be positive and determines the
	texture's depth.  In all cases, the image's height must equal the texture's
	height times its depth. */
	void load_image(const std::string &fn, int dp = -1);

	unsigned get_width() const { return width; }
	unsigned get_height() const { return height; }
	unsigned get_depth() const { return depth; }
private:
	void get_level_size(unsigned, unsigned &, unsigned &, unsigned &);

public:
	virtual AsyncLoader *load(IO::Seekable &) { return 0; }
	virtual UInt64 get_data_size() const;
	virtual void unload() { }
};

} // namespace GL
} // namespace Msp

#endif
