#include <msp/gl/extensions/arb_texture_cube_map.h>
#include <msp/io/memory.h>
#include <msp/strings/format.h>
#include "bindable.h"
#include "error.h"
#include "pixelstore.h"
#include "texturecube.h"

using namespace std;

namespace Msp {
namespace GL {

Vector3 TextureCube::directions[6] =
{
	Vector3(1, 0, 0),
	Vector3(-1, 0, 0),
	Vector3(0, 1, 0),
	Vector3(0, -1, 0),
	Vector3(0, 0, 1),
	Vector3(0, 0, -1)
};

TextureCube::TextureCube():
	Texture(GL_TEXTURE_CUBE_MAP),
	ifmt(RGB),
	size(0),
	allocated(0)
{
	static Require _req(ARB_texture_cube_map);
}

void TextureCube::storage(PixelFormat fmt, unsigned sz)
{
	if(size>0)
		throw invalid_operation("TextureCube::storage");
	if(sz==0)
		throw invalid_argument("TextureCube::storage");
	require_pixelformat(fmt);

	ifmt = fmt;
	size = sz;
}

void TextureCube::allocate(unsigned level)
{
	if(allocated&(1<<level))
		return;

	for(unsigned i=0; i<6; ++i)
		image(enumerate_faces(i), level, get_base_pixelformat(ifmt), UNSIGNED_BYTE, 0);
}

void TextureCube::image(TextureCubeFace face, unsigned level, PixelFormat fmt, DataType type, const void *data)
{
	if(size==0)
		throw invalid_operation("TextureCube::image");

	unsigned s = get_level_size(level);
	if(s==0)
		throw out_of_range("TextureCube::image");

	BindRestore _bind(this);
	glTexImage2D(face, level, ifmt, s, s, 0, fmt, type, data);

	// XXX Allocation should be tracked per-face, but we'll run out of bits
	allocated |= 1<<level;
	if(gen_mipmap && level==0)
	{
		for(; s; s>>=1, ++level) ;
		allocated |= (1<<level)-1;
	}
}

void TextureCube::image(TextureCubeFace face, const Graphics::Image &img, bool srgb)
{
	unsigned w = img.get_width();
	unsigned h = img.get_height();
	PixelFormat fmt = pixelformat_from_graphics(img.get_format());
	if(size==0)
	{
		if(w!=h)
			throw incompatible_data("TextureCube::image");

		PixelFormat f = storage_pixelformat_from_graphics(img.get_format());
		if(srgb)
			f = get_srgb_pixelformat(f);
		storage(f, w);
	}
	else if(w!=size || h!=size)
		throw incompatible_data("TextureCube::image");

	PixelStore pstore = PixelStore::from_image(img);
	BindRestore _bind_ps(pstore);

	image(face, 0, fmt, UNSIGNED_BYTE, img.get_data());
}

unsigned TextureCube::get_level_size(unsigned level)
{
	return size>>level;
}

TextureCubeFace TextureCube::enumerate_faces(unsigned i)
{
	switch(i)
	{
	case 0: return POSITIVE_X;
	case 1: return NEGATIVE_X;
	case 2: return POSITIVE_Y;
	case 3: return NEGATIVE_Y;
	case 4: return POSITIVE_Z;
	case 5: return NEGATIVE_Z;
	default: throw out_of_range("TextureCube::enumerate_faces");
	}
}

const Vector3 &TextureCube::get_face_direction(TextureCubeFace face)
{
	switch(face)
	{
	case POSITIVE_X: return directions[0];
	case NEGATIVE_X: return directions[1];
	case POSITIVE_Y: return directions[2];
	case NEGATIVE_Y: return directions[3];
	case POSITIVE_Z: return directions[4];
	case NEGATIVE_Z: return directions[5];
	default: throw invalid_argument("TextureCube::get_face_direction");
	}
}

const Vector3 &TextureCube::get_s_direction(TextureCubeFace face)
{
	switch(face)
	{
	case POSITIVE_X: return directions[5];
	case NEGATIVE_X: return directions[4];
	case POSITIVE_Y: return directions[0];
	case NEGATIVE_Y: return directions[0];
	case POSITIVE_Z: return directions[0];
	case NEGATIVE_Z: return directions[1];
	default: throw invalid_argument("TextureCube::get_s_direction");
	}
}

const Vector3 &TextureCube::get_t_direction(TextureCubeFace face)
{
	switch(face)
	{
	case POSITIVE_X: return directions[3];
	case NEGATIVE_X: return directions[3];
	case POSITIVE_Y: return directions[4];
	case NEGATIVE_Y: return directions[5];
	case POSITIVE_Z: return directions[3];
	case NEGATIVE_Z: return directions[3];
	default: throw invalid_argument("TextureCube::get_t_direction");
	}
}

Vector3 TextureCube::get_texel_direction(TextureCubeFace face, unsigned u, unsigned v)
{
	float s = (u+0.5f)*2.0f/size-1.0f;
	float t = (v+0.5f)*2.0f/size-1.0f;
	const Vector3 &fv = get_face_direction(face);
	const Vector3 &sv = get_s_direction(face);
	const Vector3 &tv = get_t_direction(face);
	return fv+s*sv+t*tv;
}

UInt64 TextureCube::get_data_size() const
{
	return id ? size*size*6*get_pixel_size(ifmt) : 0;
}


TextureCube::Loader::Loader(TextureCube &t):
	DataFile::DerivedObjectLoader<TextureCube, Texture::Loader>(t)
{
	init();
}

TextureCube::Loader::Loader(TextureCube &t, Collection &c):
	DataFile::DerivedObjectLoader<TextureCube, Texture::Loader>(t, c)
{
	init();
}

void TextureCube::Loader::init()
{
	add("image_data", &Loader::image_data);
	add("raw_data", &Loader::raw_data);
	add("storage", &Loader::storage);
}

void TextureCube::Loader::image_data(TextureCubeFace face, const string &data)
{
	Graphics::Image img;
	IO::Memory mem(data.data(), data.size());
	img.load_io(mem);

	obj.image(face, img, srgb);
}

void TextureCube::Loader::raw_data(TextureCubeFace face, const string &data)
{
	obj.image(face, 0, get_base_pixelformat(obj.ifmt), UNSIGNED_BYTE, data.data());
}

void TextureCube::Loader::storage(PixelFormat fmt, unsigned s)
{
	if(srgb)
		fmt = get_srgb_pixelformat(fmt);
	obj.storage(fmt, s);
}


void operator>>(const LexicalConverter &conv, TextureCubeFace &face)
{
	const string &str = conv.get();
	if(str=="POSITIVE_X")
		face = POSITIVE_X;
	else if(str=="NEGATIVE_X")
		face = NEGATIVE_X;
	else if(str=="POSITIVE_Y")
		face = POSITIVE_Y;
	else if(str=="NEGATIVE_Y")
		face = NEGATIVE_Y;
	else if(str=="POSITIVE_Z")
		face = POSITIVE_Z;
	else if(str=="NEGATIVE_Z")
		face = NEGATIVE_Z;
	else
		throw lexical_error(format("conversion of '%s' to TextureCubeFace", str));
}

} // namespace GL
} // namespace Msp
