#include <msp/gl/extensions/arb_multitexture.h>
#include <msp/gl/extensions/arb_vertex_shader.h>
#include "buffer.h"
#include "error.h"
#include "gl.h"
#include "mesh.h"
#include "vertexarray.h"

using namespace std;

namespace Msp {
namespace GL {

VertexArray::VertexArray(const VertexFormat &f)
{
	reset(f);
}

VertexArray::~VertexArray()
{
	/* Unbind accesses the current VertexArray, so a call from ~Bindable would
	try to access destroyed data. */
	if(current()==this)
		unbind();
}

void VertexArray::reset(const VertexFormat &f)
{
	clear();
	format = f;
	stride = get_stride(format);

	arrays.clear();

	unsigned offs = 0;
	for(const unsigned char *c=format.begin(); c!=format.end(); ++c)
	{
		unsigned slot = get_array_slot(*c);
		if(slot>=arrays.size())
			arrays.resize(slot+1);

		Array &arr = arrays[slot];
		arr.component = *c;
		arr.offset = offs;

		offs += get_component_size(*c);
	}
}

unsigned VertexArray::get_array_slot(unsigned char comp)
{
	unsigned t = get_component_type(comp);
	if(t==get_component_type(VERTEX3))
		return 0;
	else if(t==get_component_type(NORMAL3))
		return 1;
	else if(t==get_component_type(COLOR4_FLOAT))
		return 2;
	else if(comp>=TEXCOORD1 && comp<=TEXCOORD4+12)
	{
		t -= get_component_type(TEXCOORD1);
		if(t>0)
			static Require _req(ARB_multitexture);
		return 3+t;
	}
	else
	{
		static Require _req(ARB_vertex_shader);
		if(comp>=ATTRIB1)
			t -= get_component_type(ATTRIB1);
		return 7+t;
	}
}

void VertexArray::clear()
{
	data.clear();
}

void VertexArray::reserve(unsigned n)
{
	data.reserve(n*stride);
}

float *VertexArray::append()
{
	data.insert(data.end(), stride, 0.0f);
	update_offset();
	dirty = true;
	return &*(data.end()-stride);
}

float *VertexArray::modify(unsigned i)
{
	dirty = true;
	return &data[0]+i*stride;
}

unsigned VertexArray::get_data_size() const
{
	return data.size()*sizeof(float);
}

void VertexArray::bind() const
{
	if(format.empty())
		throw invalid_operation("VertexArray::apply");
	// Don't mess up the vertex array object of a mesh
	if(Mesh::current())
		throw invalid_operation("VertexArray::apply");

	const VertexArray *old = current();
	/* If the array has been modified, apply it even if it was the last one to
	be applied.  This is necessary to get the data updated to vertex buffer, and
	to resync things after a format change.  Radeon drivers also have some
	problems with modifying vertex arrays without re-setting the pointers. */
	if(!set_current(this) && !dirty)
		return;

	Buffer *vbuf = get_buffer();
	Bind _bind_vbuf(vbuf, ARRAY_BUFFER);
	if(vbuf && dirty)
		update_buffer();

	const float *base = (vbuf ? reinterpret_cast<float *>(get_offset()) : &data[0]);
	unsigned stride_bytes = stride*sizeof(float);
	apply_arrays(&arrays, (old ? &old->arrays : 0), base, stride_bytes);
}

void VertexArray::apply_arrays(const vector<Array> *arrays, const vector<Array> *old_arrays, const float *base, unsigned stride_bytes)
{
	unsigned active_tex = 0;
	unsigned n_arrays = arrays ? arrays->size() : 0;
	if(old_arrays)
		n_arrays = max<unsigned>(n_arrays, old_arrays->size());
	for(unsigned i=0; i<n_arrays; ++i)
	{
		const Array *arr = ((arrays && i<arrays->size() && (*arrays)[i].component) ? &(*arrays)[i] : 0);
		const Array *old_arr = ((old_arrays && i<old_arrays->size() && (*old_arrays)[i].component) ? &(*old_arrays)[i] : 0);
		if(!arr && !old_arr)
			continue;

		unsigned char comp = (arr ? arr->component : old_arr->component);
		unsigned sz = get_component_size(comp);
		unsigned t = get_component_type(comp);
		GLenum array_type = 0;
		if(t==get_component_type(VERTEX3))
		{
			if(arr)
				glVertexPointer(sz, GL_FLOAT, stride_bytes, base+arr->offset);
			array_type = GL_VERTEX_ARRAY;
		}
		else if(t==get_component_type(NORMAL3))
		{
			if(arr)
				glNormalPointer(GL_FLOAT, stride_bytes, base+arr->offset);
			array_type = GL_NORMAL_ARRAY;
		}
		else if(t==get_component_type(COLOR4_FLOAT))
		{
			if(arr)
			{
				if(sz==1)
					glColorPointer(4, GL_UNSIGNED_BYTE, stride_bytes, base+arr->offset);
				else
					glColorPointer(sz, GL_FLOAT, stride_bytes, base+arr->offset);
			}
			array_type = GL_COLOR_ARRAY;
		}
		else if(comp>=TEXCOORD1 && comp<=TEXCOORD4+12)
		{
			t -= get_component_type(TEXCOORD1);
			if(t>0 || active_tex)
			{
				glClientActiveTexture(GL_TEXTURE0+t);
				active_tex = t;
			}
			if(arr)
				glTexCoordPointer(sz, GL_FLOAT, stride_bytes, base+arr->offset);
			array_type = GL_TEXTURE_COORD_ARRAY;
		}
		else
		{
			if(t>=get_component_type(ATTRIB1))
				t -= get_component_type(ATTRIB1);
			if(arr)
				glVertexAttribPointer(t, sz, GL_FLOAT, false, stride_bytes, base+arr->offset);
		}

		// Only change enable state if needed
		if(arr && !old_arr)
		{
			if(array_type)
				glEnableClientState(array_type);
			else
				glEnableVertexAttribArray(t);
		}
		else if(old_arr && !arr)
		{
			if(array_type)
				glDisableClientState(array_type);
			else
				glDisableVertexAttribArray(t);
		}
	}

	if(active_tex)
		glClientActiveTexture(GL_TEXTURE0);
}

void VertexArray::unbind()
{
	const VertexArray *old = current();
	if(set_current(0))
		apply_arrays(0, &old->arrays, 0, 0);
}


VertexArray::Array::Array():
	component(0),
	offset(0)
{ }


VertexArray::Loader::Loader(VertexArray &a):
	VertexArrayBuilder(a)
{
	add("vertex2",   static_cast<void (Loader::*)(float, float)>(&Loader::vertex));
	add("vertex3",   static_cast<void (Loader::*)(float, float, float)>(&Loader::vertex));
	add("vertex4",   static_cast<void (Loader::*)(float, float, float, float)>(&Loader::vertex));
	add("normal3",   static_cast<void (Loader::*)(float, float, float)>(&Loader::normal));
	add("texcoord1", static_cast<void (Loader::*)(float)>(&Loader::texcoord));
	add("texcoord2", static_cast<void (Loader::*)(float, float)>(&Loader::texcoord));
	add("texcoord3", static_cast<void (Loader::*)(float, float, float)>(&Loader::texcoord));
	add("texcoord4", static_cast<void (Loader::*)(float, float, float, float)>(&Loader::texcoord));
	add("multitexcoord1", static_cast<void (Loader::*)(unsigned, float)>(&Loader::multitexcoord));
	add("multitexcoord2", static_cast<void (Loader::*)(unsigned, float, float)>(&Loader::multitexcoord));
	add("multitexcoord3", static_cast<void (Loader::*)(unsigned, float, float, float)>(&Loader::multitexcoord));
	add("multitexcoord4", static_cast<void (Loader::*)(unsigned, float, float, float, float)>(&Loader::multitexcoord));
	add("color3",    static_cast<void (Loader::*)(float, float, float)>(&Loader::color));
	add("color4",    static_cast<void (Loader::*)(float, float, float, float)>(&Loader::color));
	add("attrib1",   static_cast<void (Loader::*)(unsigned, float)>(&Loader::attrib));
	add("attrib2",   static_cast<void (Loader::*)(unsigned, float, float)>(&Loader::attrib));
	add("attrib3",   static_cast<void (Loader::*)(unsigned, float, float, float)>(&Loader::attrib));
	add("attrib4",   static_cast<void (Loader::*)(unsigned, float, float, float, float)>(&Loader::attrib));
	add("tangent3",  static_cast<void (Loader::*)(float, float, float)>(&Loader::tangent));
	add("binormal3", static_cast<void (Loader::*)(float, float, float)>(&Loader::binormal));
}

} // namespace GL
} // namespace Msp
